# -*- coding: utf-8 -*-

__LICENSE__ = """
Haworthia the Firework (VioletPy) is a free software, which is distributed 
in the hope that it will be useful, but WITHOUT ANY WARRANTY. You can 
redistribute it and/or modify it under the terms of the GNU General Public 
License as published by the Free Software Foundation. See the GNU General 
Public License for more details.
"""

__COPYRIGHT__ = """
Copyright (c) 2019-2022 by Eigenmiao. All Rights Reserved.
"""

__WEBSITE__ = """
https://github.com/liujiacode/VioletPy
"""

__VERSION__ = """
v2.4.9-x2d2s2-pre
"""

__AUTHOR__ = """
Eigenmiao (eigenmiao@outlook.com)
"""

__DATE__ = """
January 2, 2022
"""

__HELP__ = """
INTRODUCTION:
  VioletPy-Startup is the startup code for VioletPy.
  See https://eigenmiao.com/violetpy/ for more information.

VERSION:
  VioletPy-Startup {version} ({date})

AUTHOR:
  {author}

USAGE:
  violetpy [OPTION]... [FILE]

OPTION:
  -h, --help        : display this help.
  -v, --version     : output version information.
  -p, --project=DIR : project or software directory for VioletPy.

EXAMPLE:
  $> violetpy
  $> violetpy -h
  $> violetpy -l zh
  $> violetpy -p /PATH/TO/VIOLETPY/PROJECT
  $> violetpy -i /path/to/myimage.png
  $> violetpy -i /path/to/mycolors.dps
  $> violetpy -r all

PYTHON:
  >> # Use VioletPy as a Python module in code.
  >> # This code fragment could be used in other Python scripts.
  >> 
  >> from violetpy import violetpy as vp
  >> 
  >> dp_proj = "/PATH/TO/VIOLETPY/PROJECT"
  >> dp_argv = dict()
  >> dp_argv["help"] = True
  >> results = vp.run(dp_proj=dp_proj, dp_argv=dp_argv)
  >> 
  >> dp_argv = dict()
  >> dp_argv["temporary"] = True
  >> dp_argv["window"] = "0"
  >> dp_argv["input"] = "/path/to/mycolors.dps"
  >> results = vp.run(dp_proj=dp_proj, dp_argv=dp_argv)
  >> 
  >> color_set = results["colors"]
  >> selected_index = results["index"]
  >> selected_color = color_set[selected_index]
  >> 
  >> print("Name     : " + str(results["name"]))
  >> print("Hex Code : " + str(selected_color.hec))
  >> print("RGB Color: " + str(selected_color.rgb))
  >> print("HSV Color: " + str(selected_color.hsv))

NOTICE:
  For python module, option items in dictionary "dp_argv" should be full 
  names without prefix "--", such as "input".

NOTICE:
  To open this software without any option or file, please use command 
  `violetpy` or python code "vp.run()".

NOTICE:
  For other options such as "-i" and "--input", see the help information 
  of VioletPy.

COPYRIGHT:
  {copyright}
""".format(**{"version": __VERSION__[1:-1], "author": __AUTHOR__[1:-1], "copyright": __COPYRIGHT__[1:-1], "date": __DATE__[1:-1]})

__ERROR__ = """
ATTENTION PLEASE! -------------------------------------------------------------+
| Can not find the VioletPy project.                                           |
| Please use option '-p' or '--project' to specify the project directory.      |
| You can download this project from https://eigenmiao.com/violetpy/.          |
+------------------------------------------------------------------------------+
"""

import os
import sys
import time
import json
from violetpy.color import Color
from getopt import getopt


def run(dp_proj="", dp_argv={}):
    """
    Running VioletPy.
    Note that the sequence of colors in returned color set list doesn't consist with the sequence displayed in VioletPy Result window.
    The sequence displayed in VioletPy Result window is 2, 1, 0, 3, 4 according to the color set list, i.e., the first color in color list is the middle (main) color in VioletPy Result window.

    Args:
        dp_proj (str): violetpy startup argv, project directory.
        dp_argv (dict): violetpy argv, includes "help", "version", "temporary", "reset", "input", "output", "window" and (or) "sequence", such as {"help": True, "reset": "all"}.

    Returns:
        console results, contains the name (str), rule (str), create and modify times (tuple), color index (int), color set (tuple) and exec results (str) form software VioletPy.
    """

    def para_dir(path):
        """
        Find project directory at path.

        Args:
            path (str): path for project directory.

        Returns:
            Project folder (str) and name (str).
        """

        spath = str(path)

        if os.path.isfile(spath):
            path_lst = os.path.abspath(os.path.dirname(spath)).split(os.sep)

        elif os.path.isdir(spath):
            path_lst = os.path.abspath(spath).split(os.sep)

        else:
            return None, None

        # go back to root dir of VioletPy.
        if path_lst[-1] == "":
            path_lst = path_lst[:-1]

        if path_lst[-1] == "python":
            path_lst = path_lst[:-1]

        if path_lst[-1] == "main":
            path_lst = path_lst[:-1]

        if path_lst[-1] == "src":
            path_lst = path_lst[:-1]

        allow_lst = ([".", "src", "main", "python", "main.py",], [".", "VioletPy",], [".", "VioletPy.exe",])

        # perfect matching firstly.
        for file_lst in allow_lst:
            if os.path.isfile(os.sep.join(path_lst + file_lst)):
                return os.sep.join(path_lst), os.sep.join(file_lst)

        return None, None


    # load project.
    if dp_proj:
        dirname, basename = para_dir(dp_proj)

    elif os.path.isfile(os.sep.join([os.path.abspath(os.path.dirname(__file__)), "data.json"])):
        with open(os.sep.join([os.path.abspath(os.path.dirname(__file__)), "data.json"]), "r", encoding="utf-8") as f:
            try:
                data = json.load(f)

            except Exception as err:
                data = {}

        if "dp_proj" in data and data["dp_proj"]:
            dirname, basename = para_dir(data["dp_proj"])

        else:
            dirname, basename = None, None

    else:
        dirname, basename = None, None

    if not dirname or not basename:
        dirname, basename = para_dir(os.path.abspath(os.path.dirname(__file__)))

    if not dirname or not basename:
        print("\n+" + "-" * 30 + " VioletPy-Startup " + "-" * 30 + "+")
        print(__HELP__)
        print(__ERROR__)
        sys.exit()

    os.chdir(dirname)

    try:
        with open(os.sep.join([os.path.abspath(os.path.dirname(__file__)), "data.json"]), "w", encoding="utf-8") as f:
            json.dump({"dp_proj": dirname, "dp_exec": basename}, f, indent=4, ensure_ascii=False)

    except Exception as err:
        pass

    # run project.
    sdp_argv = dict()

    if isinstance(dp_argv, dict):
        sdp_argv = dp_argv

    elif isinstance(dp_argv, str):
        sdp_argv[dp_argv] = True

    elif isinstance(dp_argv, (tuple, list, set)):
        for i in dp_argv:
            sdp_argv[i] = True

    run_argv = ""

    for dp_key in sdp_argv:
        if sdp_argv[dp_key]:
            if dp_key in ("help", "version", "temporary"):
                run_argv = run_argv + " --{}".format(dp_key)

                if dp_key == "help":
                    print("\n+" + "-" * 30 + " VioletPy-Startup " + "-" * 30 + "+")
                    print(__HELP__)

                elif dp_key == "version":
                    print("VioletPy-Startup {} ({})\n".format(__VERSION__[1:-1], __DATE__[1:-1]))

            else:
                run_argv = run_argv + " --{}=\"{}\"".format(dp_key, sdp_argv[dp_key])

    if basename[-3:] == ".py":
        with open(basename, "r") as f:
            f.readline()
            f.readline()
            f.readline()
            line = f.readline()

        if line[:33] == "Haworthia the Firework (VioletPy)":
            results = os.popen("\"{}\" {}".format(sys.executable, basename + run_argv, "r"))

        else:
            print("\n+" + "-" * 30 + " VioletPy-Startup " + "-" * 30 + "+")
            print(__HELP__)
            print(__ERROR__)
            sys.exit()

    else:
        results = os.popen(basename + run_argv, "r")

    # get results.
    exec_results = results.read()

    if "help" in sdp_argv or "version" in sdp_argv:
        print(exec_results)
        exec_results = ""

    # para results.
    rt_name = "No Results Found."
    rt_rule = "Custom"
    rt_time = (time.localtime(0), time.localtime(0))
    rt_cidx = 0
    rt_cset = [Color("FFFFFF", tp="hec"), Color("FFFFFF", tp="hec"), Color("FFFFFF", tp="hec"), Color("FFFFFF", tp="hec"), Color("FFFFFF", tp="hec")]

    for line in exec_results.split("\n"):
        if len(line) > 9 and line[0] == "+":
            if line[4] in [str(i) for i in range(5)]:
                if line[2] == "*":
                    rt_cidx = int(line[4])

                try:
                    h, s, v = line[3:].split()[4:7]
                    rt_cset[int(line[4])] = Color((float(h), float(s), float(v)), tp="hsv")

                except Exception as err:
                    pass

            elif line[4:8] == "Name":
                rt_name = line[10:]

            elif line[4:8] == "Rule":
                rt_rule = line[10:].lstrip().rstrip()

                if rt_rule not in ("Analogous", "Monochromatic", "Triad", "Tetrad", "Pentad", "Complementary", "Shades", "Custom"):
                    rt_rule = "Custom"

            elif line[4:8] == "Time" and "; " in line[10:]:
                stime, mtime = line[10:].split("; ")[:2]

                try:
                    stime = time.strptime(stime.lstrip().rstrip(), "%Y-%m-%d %H:%M:%S")
                    mtime = time.strptime(mtime.lstrip().rstrip(), "%Y-%m-%d %H:%M:%S")
                    rt_time = (stime, mtime)

                except Exception as err:
                    pass

    results = {"name": str(rt_name), "rule": str(rt_rule), "times": tuple(rt_time), "index": int(rt_cidx), "colors": tuple(rt_cset)}

    return results

def startup():
    """
    Running VioletPy-Startup.
    """

    dp_proj = ""
    dp_argv = {}

    argv_opts, argv_left = getopt(sys.argv[1:], "hvtr:i:o:w:e:l:p:", ["help", "version", "temporary", "reset=", "input=", "output=", "window=", "sequence=", "project=", "lang=", "locale="])

    for opt_name,opt_value in argv_opts:
        if opt_name in ("-h", "--help"):
            dp_argv["help"] = True

        if opt_name in ("-v", "--version"):
            dp_argv["version"] = True

        if opt_name in ("-t", "--temporary"):
            dp_argv["temporary"] = True

        if opt_name in ("-r", "--reset"):
            dp_argv["reset"] = opt_value

        if opt_name in ("-i", "--input"):
            dp_argv["input"] = opt_value

        if opt_name in ("-w", "--window"):
            dp_argv["window"] = opt_value

        if opt_name in ("-e", "--sequence"):
            dp_argv["sequence"] = opt_value

        if opt_name in ("-l", "--locale", "--lang"):
            dp_argv["lang"] = opt_value

        if opt_name in ("-p", "--project"):
            dp_proj = opt_value

    if argv_left and argv_left[-1]:
        dp_argv["input"] = argv_left[-1]

    run(dp_proj=dp_proj, dp_argv=dp_argv)


if __name__ == "__main__":
    ans = startup()
